#!/usr/bin/env python

from datetime import date, datetime
from time import struct_time, gmtime
from math import acos, asin, atan2, cos, degrees, floor, fmod, pi, \
                 radians, sin, sqrt
from weakref import WeakValueDictionary

__author__ = "Terence Way"
__email__ = "terry@wayforward.net"
__version__ = '0.2.5'
MODULE = 'avdata'

import _avdata

__all__ = ['init', 'format_latitude', 'format_longitude', 'format_position',
           'NavItem', 'Airport', 'Navaid', 'Runway', 'Fix', 'Obstruction',
           'Route', 'magnetic_variation', 'nearest_by_ident',
           'nearest_by_prefix', 'nearest', 'great_circle', 'intercept']

def init(x):
    """Initialize the database, loading data given a directory or a list
    of directories.

    Returns (start_datetime, end_datetime), the effective date/time range.
    """
    s, e = _avdata.init(x)
    return datetime(*s), datetime(*e)

class NavItem(object):
    """Base class for various aeronautical data items, like airports, runways,
    navaids, waypoints, and fixes.

    _ident_: text name of item
    _position_: tuple (latitude, longitude) in radians
    """
    def __init__(self, id, ident, position):
        self.id = id
        self.ident = ident
        self.position = position

    def __str__(self):
        return '%s %s' % (self.ident, format_position(self.position))

    def magnetic_variation(self):
        return magnetic_variation(self.position)

    def routes(self):
        """Returns a list of route legs starting from this fix.

        Each leg is a tuple (avdata.Route, avdata.Fix)
        """
        return _avdata.routes(self.id)

    def route(self, route_name, end):
        """Returns [list of fixes]) given a route name and an endpoint.
        The endpoint can be an intersecting route, as long as there's a
        shared intersection or fix."""

        if route_name == 'DIRECT':
            next = self.nearest_by_ident(end)
            return (next or []) and [next]

        if len(route_name) > 3 and route_name[3:].isdigit():
            # possibly VOR and radial?
            vor = self.nearest_by_ident(route_name[:3])
            rad0 = int(route_name[3:])

            if isinstance(vor, Navaid) and 0 < rad0 <= 360:
                # definitely VOR and radial
                next = self.nearest_by_ident(end)
                if not next:
                    return []

                d, a = great_circle(vor.position, next.position)
                rad1 = fmod(degrees(a + vor.magvar) + 360.0, 360.0)
                if (-1.5 < (rad1 - rad0) < 1.5):
                    return [next]

        queue = [self]
        found = {self.id: None}

        while queue:
            t = queue.pop()
            if t.ident == end or (isinstance(t, Navaid) and t.closest().ident == end):
                return _fix_list(found, t)
            a = _avdata.routes(t.id)
            for r, fix in a:
                if r.ident == end:
                    return _fix_list(found, t)
                if r.ident == route_name:
                    if fix.id not in found:
                        found[fix.id] = t
                        queue.append(fix)

        return []

    def format_position(self, deg=u'\xb0'):
        """Returns a string in 'Ndd mm.s Wdd mm.s' format.  The _deg_ keyword
        argument replaces the degree mark.

        Example:
        >> a.format_position()
        u'N19\xb0 43.2 W155\xb0 02.9'

        >> a.format_position(deg='deg')
        'N19deg 43.2 W155deg 02.9'
        """
        return format_position(self.position, deg)

    def format_latitude(self, deg=u'\xb0'):
        """Returns the latitude as a human-readable string in 'Ndd mms'
        format.  The _deg_ keyword argument replaces the degree mark.

        Example:
        >> a.format_latitude()
        u'N19\xb0 43.2'

        >> a.format_latitude(deg='&deg;')
        'N19&deg; 43.2'
        """
        return format_latitude(self.position[0], deg)

    def format_longitude(self, deg=u'\xb0'):
        """Returns the longitude as a human-readable string in 'Eddd mms'
        format.  The _deg_ keyword argument replaces the degree mark.

        Example:
        >> a.format_longitude()
        u'W155\xb0 02.9'

        >> a.format_longitude(deg='&deg;')
        'W155&deg; 02.9'
        """
        return format_longitude(self.position[1], deg)

    def nearest_by_ident(self, ident):
        return nearest_by_ident(ident, self.position)

    def nearest_by_prefix(self, prefix):
        return nearest_by_prefix(prefix, self.position)

    def nearest(self, range):
        return nearest(self.position, range)

    def closest(self, type=None, range=10):
        type = type or Airport
        result = None
        for i in self.nearest(range):
            if isinstance(i, type):
                d = self.distance_to(i.position)
                if d < range:
                    result = i
                    range = d
        return result

    def extend(self, tc, distance):
        """Extend a radial/distance, returning (lat, lon)
        tc is course/radial in true radians
        distance is in nautical miles
        """
        lat1, lon1 = self.position
        tc = -tc # why!?
        d = (pi / 10800.0) * distance
        lat = asin(sin(lat1) * cos(d) + cos(lat1) * sin(d) * cos(tc))
        if cos(lat) == 0.0:
            lon = lon1
        else:
            lon = fmod(lon1 - asin(sin(tc) * sin(d) / cos(lat)) + pi,
                       2.0 * pi) - pi

        return lat, lon

    def great_circle(self, next):
        if isinstance(next, NavItem):
            next = next.position

        return great_circle(self.position, next)

    def distance_to(self, next):
        return self.great_circle(next)[0]

    def course_to(self, next):
        return self.great_circle(next)[1]

def _fix_list(d, f):
    result = []
    while f:
        result.append(f)
        f = d[f.id]
    result.reverse()
    return result[1:]

def format_position((latitude, longitude), deg=''):
    """Given a tuple (latitude, longitude), in radians, returns a string
    with latitude, longitude in degrees

    >>> format_position((1, 2))
    'N57 17.7 E114 35.5'

    >>> format_position((1, 2), deg='&deg;')
    'N57&deg; 17.7 E114&deg; 35.5'
    """
    return '%s %s' % (format_latitude(latitude, deg),
                      format_longitude(longitude, deg))

def format_latitude(latitude, deg=''):
    """Given a latitude, in radians, returns a string in the format
    'Ndd mm.s'.

    >>> format_latitude(1)
    'N57 17.7'

    >>> format_latitude(-1)
    'S57 17.7'

    >>> format_latitude(1, deg='&deg;')
    'N57&deg; 17.7'
    """
    return _format_radians(latitude, 'NS', deg)

def format_longitude(longitude, deg=''):
    """Given a longitude, in radians, returns a string in the format
    'Edd mm.s'.

    >>> format_longitude(2)
    'E114 35.5'

    >>> format_longitude(-2)
    'W114 35.5'

    >>> format_longitude(-2, deg='&deg;')
    'W114&deg; 35.5'
    """
    return _format_radians(longitude, 'EW', deg)

def _format_radians(r, s, deg=''):
    """Format radians as '[NSEW]DD_ MM.M', with _ replaced by the _deg_
    parameter.

    Examples:
    >>> _format_radians(radians(10), 'NS')
    'N10 00.0'

    >>> _format_radians(radians(9.999), 'NS', deg='&deg;')
    'N9&deg; 59.9'

    >>> _format_radians(radians(9.9999), 'NS')
    'N10 00.0'

    >>> _format_radians(-radians(9.999), 'NS', deg='&deg;')
    'S9&deg; 59.9'

    >>> _format_radians(-radians(9.9999), 'NS')
    'S10 00.0'

    >>> _format_radians(radians(-153.242), 'EW')
    'W153 14.5'
    """
    r = int(round(degrees(r) * 600))

    if r < 0:
        prefix = s[1]
        r = -r
    else:
        prefix = s[0]

    d, m = divmod(r, 600)
    str = '%s%d%s %03d' % (prefix, d, deg, m)
    return str[:-1] + '.' + str[-1]

class Airport(NavItem):
    """Represents an airport.

    _ident_: Unique 3- or 4- letter identifier, usually based on ICAO ident
    _position_: Tuple (latitude, longitude) in radians
    _elevation_: in feet msl
    """
    def __init__(self, id, ident, position, elevation):
        NavItem.__init__(self, id, ident, position)
        self.elevation = elevation

    def __repr__(self):
        return 'avdata.Airport(%d, "%s", (%f, %f), %d)' % (self.id, self.ident,
		self.position[0], self.position[1], self.elevation)

    def runways(self):
        return [x for x in self.nearest(5) \
                if isinstance(x, Runway) and x.ident.startswith(self.ident)]

class Runway(NavItem):
    """Represents a runway.
    """
    def __init__(self, id, ident, position, tdze, length, width):
        NavItem.__init__(self, id, ident, position)
        self.tdze = tdze
        self.length = length
        self.width = width

    def __repr__(self):
        return 'avdata.Runway(%d, "%s", (%f, %f), %d)' % (self.id, self.ident,
		self.position[0], self.position[1], self.tdze)

    def airport(self):
        """Returns an Airport object for this runway."""
        return self.nearest_by_ident(self.ident.split('-', 1)[0])

class Navaid(NavItem):
    """Represents a navaid: a VOR, an NDB, localizer, etc.

    _ident_: not unique, string identifier for this navaid
    _position_: tuple (latitude, longitude) in radians
    _elevation_: in feet msl
    _frequency_: string, context specific (may be in MHz, KHz)
    _magvar_: magnetic variation of record, necessary for VORs
    """
    def __init__(self, id, ident, position, elevation, frequency, magvar):
        NavItem.__init__(self, id, ident, position)
        self.elevation = elevation
        self.frequency = frequency
        self.magvar = magvar

    def __repr__(self):
        return 'avdata.Navaid(%d, "%s", (%f, %f), %d, "%s", %f)' % (self.id,
		self.ident, self.position[0], self.position[1], self.elevation,
		self.frequency, self.magvar)

    def magnetic_variation(self):
        return self.magvar

class Fix(NavItem):
    def __repr__(self):
        return 'avdata.Fix(%d, "%s", (%f, %f))' % (self.id, self.ident,
		self.position[0], self.position[1])

class Obstruction(object):
    """A tower, crane, building, or other obstruction.

    _position_: (latitude, longitude) tuple in radians
    _msl_: height in feet, msl
    _agl_: height in feet, agl
    """
    def __init__(self, id, position, msl, agl):
        self.id = id
        self.position = position
        self.msl = msl
        self.agl = agl

    def __str__(self):
        return '/\ %s %d(%d)' % (format_position(self.position), self.msl, self.agl)

    def __repr__(self):
        return 'avdata.Obstruction(%d, (%f, %f), %d, %d)' % (self.id,
		self.position[0], self.position[1], self.msl, self.agl)

class Route(object):
    DP, STAR, VICTOR, JET, AMBER, GREEN, BLUE, RED, RNAV, IAP, \
    DIRECT, INTERCEPT, VECTOR, RADIAL = range(14)

    types = ['DP', 'STAR', 'VICTOR', 'JET', 'AMBER', 'GREEN', 'BLUE', 'RED',
             'RNAV', 'IAP', 'DIRECT', 'INTERCEPT', 'VECTOR', 'RADIAL']

    def __init__(self, id, ident, type):
        self.id = id
        self.ident = ident
        self.type = type

    def __repr__(self):
        return 'avdata.Route(%d, "%s", avdata.Route.%s)' % (self.id,
                    self.ident, Route.types[self.type])

def magnetic_variation(position, dt=None, msl=0):
    """Returns magnetic variation, in radians, given a position and a datetime.

    _dt_ defaults to today's date.  It can be given in time.struct_time,
    datetime.date or datetime.datetime, or as a decimal value (2010.9).

    _msl_ defaults to 0.  Meters MSL.

    Returns a negative value for east variation, positive for west variation
    """
    return _avdata.magnetic_variation(decimal_date(dt), position, msl)

def decimal_date(dt=None):
    if isinstance(dt, date) or isinstance(dt, datetime):
        dt = dt.timetuple()
    else:
        dt = dt or gmtime()

    if isinstance(dt, struct_time):
        y = dt.tm_year
        if y % 4 == 0 and (y % 100 != 0 or y % 400 == 0):
            return y + dt.tm_yday / 366.0
        else:
            return y + dt.tm_yday / 365.0
    else:
        return float(dt)

def nearest_by_ident(ident, position):
    if '/' in ident:
        lat, lon = ident.split('/', 1)
        lat = parse_degrees(lat, 'N', 'S')
        lon = parse_degrees(lon, 'E', 'W')
        return Fix(0, ident, (lat, lon))

    if len(ident) > 7 and ident[-6:].isdigit():
        # FIXRRRDDD: ex ITO270020
        radial = float(ident[-6:-3])
        dme = float(ident[-3:])
        name = ident[:-6]
        fix = _avdata.nearest_by_ident(name, position)

        if fix:
            return Fix(0, ident, fix.extend(radians(radial) - fix.magnetic_variation(), dme))
    else:
        return _avdata.nearest_by_ident(ident, position)

def parse_degrees(str, pos, neg):
    if str[0] in pos or str[0] == '+':
        sign = 1.0
        s = str[1:]
    elif str[0] in neg or str[0] == '-':
        sign = -1.0
        s = str[1:]
    elif str[-1] in pos:
        sign = 1.0
        s = str[:-1]
    elif str[-1] in neg:
        sign = -1.0
        s = str[:-1]
    else:
        sign = 1.0
        s = str

    if '.' in s:
        s, tenths = s.split('.', 1)
    else:
        tenths = '0'

    x = s.count(':')
    l = len(s)
    if x == 1:
        deg, min = s.split(':')
        sec = 0.0
    elif x == 2:
        deg, min, sec = s.split(':')
    elif x > 2:
        raise Exception("cannot parse latitude/longitude " + str)
    elif l < 4:
        deg = s
        min = sec = 0.0
    elif l < 6:
        deg = s[:-2]
        min = s[-2:]
        sec = 0.0
    else:
        sec = s[-2:] + '.' + tenths
        min = s[-4:-2]
        deg = s[:-4]

    return sign * radians(int(deg) + int(min) / 60.0 + float(sec) / 3600.0)

nearest_by_prefix = _avdata.nearest_by_prefix
nearest = _avdata.nearest

_cache = WeakValueDictionary()

def cache(type, id, *args):
    if id in _cache:
        value = _cache[id]
        assert isinstance(value, type)
    else:
        value = type(id, *args)
        _cache[id] = value
    return value

# A brief note about rounding and accuracy: we round to the most
# conservative estimate for fuel consumption.  For example, when
# calculating speeds, we round DOWN to the nearest knot, when
# calculating distances, we round UP to the nearest 1/10 nautical mile.

def great_circle(pos1, pos2):
    """Calculate distance (in nautical miles) and course (in radians)
    between two points on the Earth, using a great circle path.

    Returns (nm, radians).

    For example, what is the great circle path from LAX (33 degrees,
    57 minutes North, 118 degrees, 24 minutes West) to JFK (40
    degrees 38 minutes North, 73 degrees 47 minutes West)?

    >>> lax = (radians(33 + 57 / 60.0), -radians(118 + 24 / 60.0))
    >>> jfk = (radians(40 + 38 / 60.0), -radians(73 + 47 / 60.0))

    The distance is 2143.7 nautical miles:
    >>> round(great_circle(lax, jfk)[0], 1)
    2143.7

    The initial true course out of LAX is 66 degrees:
    >>> round(degrees(great_circle(lax, jfk)[1]), 1)
    65.9

    pre:
        -pi <= lat1 <= pi
        -pi <= lat2 <= pi
        -pi <= long1 <= 2 * pi
        -pi <= long2 <= 2 * pi
    """
    # of two formula, this one is less susceptible to rounding errors
    # for short distances
    #
    lat1, long1 = pos1
    lat2, long2 = pos2
    cos_lat1 = cos(lat1)

    d = 2 * asin(sqrt(sin((lat1 - lat2) / 2.0) ** 2 + \
                      cos_lat1 * cos(lat2) * sin((long2 - long1) / 2.0) ** 2))

    # Short circuit if the same point
    if d == 0.0: return (0.0, 0.0)

    tc1 = (sin(lat2) - sin(lat1) * cos(d)) / (sin(d) * cos_lat1)
    tc1 = acos(max(min(tc1, 1.0), -1.0))

    if sin(long1 - long2) >= 0:
        tc1 = 2 * pi - tc1

    return 10800.0 * d / pi, tc1

def intersect(v1, v2):
    lat1, lon1, r1 = v1
    lat2, lon2, r2 = v2

    dst12 = 2 * asin(sqrt((sin((lat1-lat2) / 2)) ** 2 + \
            cos(lat1) * cos(lat2) * sin( (lon1-lon2) / 2) ** 2))

    if sin(lon2-lon1) < 0.0:
        crs12 = acos((sin(lat2) - sin(lat1) * cos(dst12)) / \
                (sin(dst12) * cos(lat1)))
        crs21 = 2.0 * pi - acos((sin(lat1) - sin(lat2) * cos(dst12)) / \
                (sin(dst12) * cos(lat2)))
    else:
        crs12 = 2.0 * pi - acos((sin(lat2) - sin(lat1) * cos(dst12)) / \
                (sin(dst12) * cos(lat1)))
        crs21 = acos((sin(lat1) - sin(lat2) * cos(dst12)) / \
                (sin(dst12) * cos(lat2)))

    ang1 = fmod(r1 - crs12 + pi, 2.0*pi) - pi
    ang2 = fmod(crs21 - r2 + pi, 2.0*pi) - pi

    if sin(ang1) == 0 and sin(ang2) == 0:
        raise Exception("infinity of intersections")

    if sin(ang1) * sin(ang2) < 0:
        raise Exception("intersection ambiguous")

    ang1 = abs(ang1)
    ang2 = abs(ang2)
    ang3 = acos(-cos(ang1) * cos(ang2) + sin(ang1) * sin(ang2) * cos(dst12)) 
    dst13 = atan2(sin(dst12) * sin(ang1) * sin(ang2),
                  cos(ang2) + cos(ang1) * cos(ang3))
    lat3 = asin(sin(lat1) * cos(dst13) + cos(lat1) * sin(dst13) * cos(r1))
    dlon = atan2(sin(r1) * sin(dst13) * cos(lat1),
                 cos(dst13) - sin(lat1) * sin(lat3))
    lon3 = fmod(lon1 - dlon + pi, 2*pi) - pi

    return lat3, lon3

def _test():
    import doctest, avdata
    return doctest.testmod(avdata)

if __name__ == '__main__':
    _test()
